@@ -60,6 +60,7 @@ public:
 
 namespace mold {
 
+namespace ranges = std::ranges;
 using namespace std::literals::string_literals;
 using namespace std::literals::string_view_literals;
 
@@ -68,8 +69,6 @@ inline char *output_tmpfile;
 inline u8 *output_buffer_start = nullptr;
 inline u8 *output_buffer_end = nullptr;
 
-extern std::string mold_git_hash;
-
 std::string errno_string();
 void cleanup();
 void install_signal_handler();
@@ -416,14 +415,6 @@ inline std::vector<T> flatten(std::vector<std::vector<T>> &vec) {
   return ret;
 }
 
-inline void sort(auto &vec) {
-  std::stable_sort(vec.begin(), vec.end());
-}
-
-inline void sort(auto &vec, auto less) {
-  std::stable_sort(vec.begin(), vec.end(), less);
-}
-
 template <typename T>
 inline void remove_duplicates(std::vector<T> &vec) {
   vec.erase(std::unique(vec.begin(), vec.end()), vec.end());
@@ -490,6 +481,23 @@ inline u64 read_uleb(u8 *buf) {
   return read_uleb(&tmp);
 }
 
+inline i64 read_sleb(u8 **buf) {
+  u64 val = 0;
+  u8 shift = 0;
+  u8 byte;
+  do {
+    byte = *(*buf)++;
+    val |= (byte & 0x7f) << shift;
+    shift += 7;
+  } while (byte & 0x80);
+  return sign_extend(val, shift);
+}
+
+inline i64 read_sleb(u8 *buf) {
+  u8 *tmp = buf;
+  return read_sleb(&tmp);
+}
+
 inline u64 read_uleb(std::string_view *str) {
   u8 *start = (u8 *)str->data();
   u8 *ptr = start;
@@ -657,7 +665,7 @@ public:
 
     // Since the shard is circular, we need to handle the last entries
     // as if they were next to the first entries.
-    while (entries[end - 1].key)
+    while (begin < end && entries[end - 1].key)
       vec.push_back(entries + --end);
 
     // Find entries contiguous in the buckets and sort them.
